"""Perform database migrations using Alembic.

This is a click wrapper around the alembic library.
"""
import logging
import os

from flask import current_app
from flask.cli import AppGroup
import click
import alembic
import alembic.config

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

DEFAULT_DIR = 'migrations'
EXT_NAME = 'migrate'
HELP = {
	'autogenerate': (
		'Populate revision script with candidate migration '
		'operations, based on comparison of database to model.'
		),
	'branch-label': 'Specify a branch label to apply to the new revision',
	'head': 'Specify head revision or <branchname>@head to base new revision on.',
	'message': "Message string to use with 'revision'",
	'resolve-dependencies': 'Treat dependency versions as down revisions',
	'rev-id': 'Specify a hardcoded revision id instead of generating one',
	'rev-range': 'Specify a revision range; format is [start]:[end]',
	'splice': 'Allow a non-head revision as the "head" to splice onto',
	'sql': "Don't emit SQL to database - dump to standard output/file instead",
	'tag': "Arbitrary 'tag' name - can be used by custom env.py scripts.",
	'verbose': 'Use more verbose output',
	'version-path': 'Specify specific path from config for version file',
	'x-arg': "Additional arguments consumed by custom env.py scripts",
	}


class _MigrateConfig(object):
	def __init__(self, directory, db, **kwargs):
		self.directory = directory
		self.db = db
		self.configure_args = kwargs


class Migrate(object):
	"""Flask extension for alembic migrations."""
	def __init__(self, app=None, db=None, directory=DEFAULT_DIR, **kwargs):
		if app is not None:
			self.init_app(app, db, directory, **kwargs)

	def init_app(self, app, db=None, directory=DEFAULT_DIR, **kwargs):
		if not hasattr(app, 'extensions'):
			app.extensions = {}
		app.extensions[EXT_NAME] = _MigrateConfig(directory, db, **kwargs)


class Config(alembic.config.Config):
	def get_template_directory(self):
		package_dir = os.path.abspath(os.path.dirname(__file__))
		return os.path.join(package_dir, 'templates')


def _get_config(x_arg=None):
	directory = current_app.extensions[EXT_NAME].directory
	config = Config(os.path.join(directory, 'alembic.ini'))
	config.set_main_option('script_location', directory)
	if x_arg is not None:
		if config.cmd_opts is None:
			config.cmd_opts = lambda: None
		if not getattr(config.cmd_opts, 'x', None):
			config.cmd_opts.x = []
		config.cmd_opts.x.append(x_arg)
	return config


cli = AppGroup(help=__doc__, name='migrate')


@cli.command()
def init():
	"""Generates a new migration"""
	directory = current_app.extensions[EXT_NAME].directory
	config = Config()
	config.set_main_option('script_location', directory)
	config.config_file_name = os.path.join(directory, 'alembic.ini')
	alembic.command.init(config, directory, 'flask')


@cli.command()
@click.option('-m', '--message', help=HELP['message'])
@click.option(
	'--autogenerate/--no-autogenerate',
	default=True,
	help=HELP['autogenerate'],
	)
@click.option('--sql', is_flag=True, help=HELP['sql'])
@click.option('--head', default='head', help=HELP['head'])
@click.option('--splice', is_flag=True, help=HELP['splice'])
@click.option('--branch-label', help=HELP['branch-label'])
@click.option('--version-path', help=HELP['version-path'])
@click.option('--rev-id', help=HELP['rev-id'])
def revision(
	message,
	autogenerate,
	sql,
	head,
	splice,
	branch_label,
	version_path,
	rev_id,
	):
	"""Create a new revision file."""
	config = _get_config()
	alembic.command.revision(
		config,
		message,
		autogenerate=autogenerate,
		sql=sql,
		head=head,
		splice=splice,
		branch_label=branch_label,
		version_path=version_path,
		rev_id=rev_id,
		)


@cli.command()
@click.argument('revisions', nargs=-1)
@click.option('-m', '--message', help=HELP['message'])
@click.option('--branch-label', help=HELP['branch-label'])
@click.option('--rev-id', help=HELP['rev-id'])
def merge(revisions, message, branch_label, rev_id):
	"""Merge two revisions together.  Creates a new migration file.

	revisions is a list of one or more revisions, or 'heads' for all heads.
	"""
	config = _get_config()
	alembic.command.merge(
		config,
		revisions,
		message=message,
		branch_label=branch_label,
		rev_id=rev_id,
		)


@cli.command()
@click.argument('revision', required=False, default='head')
@click.option('--tag', help=HELP['tag'])
@click.option('--sql', is_flag=True, help=HELP['sql'])
@click.option('-x', '--x-arg', help=HELP['x-arg'])
def upgrade(revision, sql, tag, x_arg):
	"""Upgrade to a later version.

	revision is the identifier of the revision to upgrade to.
	"""
	config = _get_config(x_arg=x_arg)
	alembic.command.upgrade(config, revision, sql=sql, tag=tag)


@cli.command()
@click.argument('revision', required=False, default='-1')
@click.option('--tag', help=HELP['tag'])
@click.option('--sql', is_flag=True, help=HELP['sql'])
@click.option('-x', '--x-arg', help=HELP['x-arg'])
def downgrade(revision, sql, tag, x_arg):
	"""Revert to a previous version

	revision is the identifier of the revision to downgrade to.
	"""
	config = _get_config(x_arg=x_arg)
	if sql and revision == '-1':
		revision = 'head:-1'
	alembic.command.downgrade(config, revision, sql=sql, tag=tag)


@cli.command()
@click.argument('revision', required=False, default='head')
def show(revision):
	"""Show the revision denoted by the given symbol."""
	config = _get_config()
	alembic.command.show(config, revision)


@cli.command()
@click.option('-v', '--verbose', is_flag=True, help=HELP['verbose'])
@click.option('-r', '--rev-range', help=HELP['rev-range'])
def history(rev_range, verbose):
	"""List changeset scripts in chronological order."""
	config = _get_config()
	alembic.command.history(config, rev_range, verbose=verbose)


@cli.command()
@click.option(
	'--resolve-dependencies',
	is_flag=True,
	help=HELP['resolve-dependencies'],
	)
@click.option('-v', '--verbose', is_flag=True, help=HELP['verbose'])
def heads(verbose, resolve_dependencies):
	"""Show current available heads in the script directory"""
	config = _get_config()
	alembic.command.heads(
		config,
		verbose=verbose,
		resolve_dependencies=resolve_dependencies,
		)


@cli.command()
@click.option('-v', '--verbose', is_flag=True, help=HELP['verbose'])
def branches(verbose):
	"""Show current branch points"""
	config = _get_config()
	alembic.command.branches(config, verbose=verbose)


@cli.command()
@click.option('-v', '--verbose', is_flag=True, help=HELP['verbose'])
def current(verbose, head_only):
	"""Display the current revision for each database."""
	config = _get_config()
	alembic.command.current(config, verbose=verbose)


@cli.command()
@click.option('--tag', help=HELP['tag'])
@click.option('--sql', is_flag=True, help=HELP['sql'])
@click.argument('revision', required=False, default='head')
def stamp(revision, sql, tag):
	"""'stamp' the revision table with the given revision; don't run any
	migrations"""
	config = _get_config()
	alembic.command.stamp(config, revision, sql=sql, tag=tag)



"""Utils for working with SQLAlchemy."""
import csv
from decimal import Decimal
import fractions
import logging
import os.path

from flask import abort
import flask_sqlalchemy
from sqlalchemy import (
	types,
	Column,
	Integer,
	String,
	ForeignKey,
	)
from sqlalchemy.inspection import inspect
from sqlalchemy.orm.exc import MultipleResultsFound, NoResultFound
from sqlalchemy.orm.collections import (
	InstrumentedList,
	InstrumentedSet,
	InstrumentedDict,
	)
import pytz

logger = logging.getLogger(__name__)
db = flask_sqlalchemy.SQLAlchemy()


def psycopg_uri(username, password, db_name, host='', port=None):
	"""Create a URL for psycopg2.

	http://docs.sqlalchemy.org/en/latest/dialects/postgresql.html#module-sqlalchemy.dialects.postgresql.psycopg2  # noqa
	"""
	if port:
		return f'postgresql+psycopg2://{username}:{password}@{host}:{port}/{db_name}'
	else:
		return f'postgresql+psycopg2://{username}:{password}@{host}/{db_name}'


def foreign_key_col(col, **kwargs):
	return Column(
		col.type,
		ForeignKey(col, ondelete='CASCADE', onupdate='CASCADE'),
		**kwargs
		)


def parent_key(column, col_type=Integer, nullable=False, index=True, **kwargs):
	return Column(
		col_type,
		ForeignKey(column, ondelete='CASCADE', onupdate='CASCADE'),
		nullable=nullable,
		index=index,
		**kwargs
		)


class Fraction(types.TypeDecorator):
	"""Type for storing and retrieving Fractions.

	Currently, this is backed by a Decimal, so some precision may be lost on
	conversion.
	"""

	impl = types.Numeric

	def process_bind_param(self, value, dialect):
		if value is None:
			return None
		assert isinstance(value, int) or isinstance(value, fractions.Fraction)
		return value.numerator / Decimal(value.denominator)

	def process_result_value(self, value, dialect):
		if value is None:
			return None
		assert isinstance(value, Decimal)
		return fractions.Fraction(value).limit_denominator()


class UTCDateTime(types.TypeDecorator):
	"""Type for storing and retrieving DateTimes as UTC.

	Naive datetimes are assumed to be UTC.
	The returned value is always a non-naive UTC datetime.
	"""

	impl = types.TIMESTAMP(timezone=True)

	def process_bind_param(self, value, dialect):
		if value is None:
			return None
		if value.tzinfo is None:
			raise ValueError('Cannot handle naive datetime')
		return value.astimezone(pytz.utc)

	def process_result_value(self, value, dialect):
		if value is None:
			return None
		return value.astimezone(pytz.utc)


class BaseMixin():
	@classmethod
	def find_one(cls, **kwargs):
		"""Query this table for a single row matching kwargs filters."""
		return cls.query.filter_by(**kwargs).one()

	@classmethod
	def find_one_or_404(cls, **kwargs):
		"""Query this table for a single row, flask.abort(404) if not found."""
		try:
			cls.find_one(**kwargs)
		except (NoResultFound, MultipleResultsFound):
			abort(404)

	@classmethod
	def create(cls, **kwargs):
		obj = cls(**kwargs)
		cls.query.session.add(obj)
		return obj

	@classmethod
	def find_create(cls, create_args=None, **kwargs):
		"""Find or create an instance of this model.

		Optionally provide arguments used only for creating the object, not
		querying.
		"""
		try:
			return cls.find_one(**kwargs)
		except NoResultFound:
			create_args = dict(create_args or {})
			create_args.update(kwargs)
			return cls.create(**create_args)

	@classmethod
	def exists(cls, **kwargs):
		try:
			cls.find_one(**kwargs)
		except NoResultFound:
			return False
		else:
			return True

	@classmethod
	def cast(cls, obj, allow_none=True):
		if obj is None:
			if allow_none:
				return None
			else:
				raise ValueError(f'Cannot cast None to {cls}')
		elif isinstance(obj, cls):
			return obj
		else:
			raise TypeError(f'Cannnot cast an object of {type(obj)} to {cls}')

	@classmethod
	def _repr_class_template(cls):
		col_names = [col.name for col in cls.__table__.columns]
		item_format = '{col}={{self.{col}!r}}'
		fields = ', '.join(item_format.format(col=col) for col in col_names)
		return '{class_name}({fields})'.format(
			class_name=cls.__name__,
			fields=fields,
			)

	def __repr__(self):
		class_template = self._repr_class_template()
		return class_template.format(obj=self, self=self)

	@classmethod
	def query_default_order(cls):
		"""Return a class query with a default order."""
		raise NotImplementedError

	@classmethod
	def _get_pkey_col(cls):
		primary_key_cols = inspect(cls).primary_key
		if len(primary_key_cols) != 1:
			msg = 'Class %s does not have exactly one primary key column' % cls
			raise NotImplementedError(msg)
		return primary_key_cols[0]

	@classmethod
	def fkey_constraint(cls, *args, ondelete='CASCADE', onupdate='CASCADE'):
		"""Return a ForeignKeyConstraint for the primary keys of this model."""
		pkey_col = cls._get_pkey_col()
		return ForeignKey(pkey_col, *args, ondelete=ondelete, onupdate=onupdate)

	@classmethod
	def pkey(cls, **kwargs):
		"""Return a Column definition for the primary key of this model."""
		pkey_col = cls._get_pkey_col()
		return foreign_key_col(pkey_col, **kwargs)

	def to_dict(self, sub=None):
		"""Create a dict of this obj's attributes and optionally related objects.

		Specify which relationships to include as keys in `sub`. The value of
		each key is the argument to pass to that relation's `to_dict`.

		For example:

		class Character(BaseModel):
			first_name = Column(String)
			last_name = Column(String)
			parents = relationship(Character)
			siblings = relationship(Character)

		`bart.to_dict({'parents': {'siblings': None}}) == {
			'first_name': 'Bart',
			'last_name': 'Simpson',
			'parents': [
				{
					'first_name': 'Homer',
					'last_name': 'Simpson',
					'siblings': [],
					},
				{
					'first_name': 'Marge',
					'last_name': 'Simpson',
					'siblings': [
						{
							'first_name': 'Patty',
							'last_name': 'Bouvier',
							},
						{
							'first_name': 'Selma',
							'last_name': 'Bouvier',
							},
						],
					},
				]
			}`
		"""
		instance_state = inspect(self)
		columns = instance_state.mapper.column_attrs
		out = {col.key: getattr(self, col.key) for col in columns}
		sub = sub or {}
		for relationship_name, args in sub.items():
			relationship = getattr(self, relationship_name)
			try:
				out[relationship_name] = relationship.to_dict(args)
			except AttributeError:
				if isinstance(relationship, (InstrumentedList, InstrumentedSet)):
					out[relationship_name] = []
					for relation in relationship:
						out[relationship_name].append(relation.to_dict(args))
				elif isinstance(relationship, InstrumentedDict):
					out[relationship_name] = {}
					for k, relation in relationship.items():
						out[relationship_name][k] = relation.to_dict(args)
				else:
					msg = "Don't know how to handle relationship of type {rel_type}".format(
						rel_type=type(relationship),
						)
					raise NotImplementedError(msg)
		return out

	@classmethod
	def _get_path(cls, path, directory):
		if path is None:
			path = '%s.csv' % cls.__tablename__
			if directory:
				path = os.path.join(directory, path)
		else:
			if directory:
				raise ValueError('Must not pass path and directory')
		return path

	@classmethod
	def merge_csv(cls, path=None, directory=None, io_wrapper=None):
		"""Load data from a csv file and merge into db.

		This is slower than import because orm objects are created.
		"""
		path = cls._get_path(path, directory)
		logger.info('Importing %s to %s', path, cls)
		with open(path, 'r', newline='') as infile:
			cls.merge_from_file(infile, io_wrapper=io_wrapper)

	@classmethod
	def import_csv(cls, path=None, directory=None, io_wrapper=None):
		"""Import data from a csv file."""
		path = cls._get_path(path, directory)
		logger.info('Importing %s to %s', path, cls)
		with open(path, 'r', newline='') as infile:
			cls.import_from_file(infile, io_wrapper=io_wrapper)

	@classmethod
	def export_csv(cls, path=None, directory=None, io_wrapper=None):
		"""Export data to csv file."""
		path = cls._get_path(path, directory)
		logger.info('Exporting %s to %s', cls, path)
		try:
			with open(path, 'r', newline='') as infile:
				reader = csv.DictReader(infile)
				fieldnames = reader.fieldnames
		except IOError:
			fieldnames = None
		with open(path, 'w', newline='') as outfile:
			cls.export_to_file(outfile, fieldnames=fieldnames, io_wrapper=io_wrapper)

	@classmethod
	def import_from_file(cls, infile, io_wrapper=None):
		logger.info('Importing to %s', cls)
		if io_wrapper:
			infile = io_wrapper(infile)
		reader = csv.DictReader(infile)
		# TODO try insert many (not on session)
		# http://stackoverflow.com/questions/25694234/bulk-update-in-sqlalchemy-core-using-where
		# db.session.bulk_insert_mappings(cls, reader)
		rows = list(reader)
		db.engine.execute(cls.__table__.insert(), rows)

	@classmethod
	def merge_from_file(cls, infile, io_wrapper=None):
		logger.info('Merging to %s', cls)
		if io_wrapper:
			infile = io_wrapper(infile)
		reader = csv.DictReader(infile)
		for row in reader:
			obj = cls(**row)
			cls.query.session.merge(obj)

	@classmethod
	def export_to_file(cls, outfile, fieldnames=None, io_wrapper=None):
		logger.info('Exporting %s', cls)
		if io_wrapper:
			outfile = io_wrapper(outfile)
		fieldnames = fieldnames or []
		for col in cls.__table__.columns:
			if col.name not in fieldnames:
				fieldnames.append(col.name)
		writer = csv.DictWriter(outfile, fieldnames, extrasaction='ignore')
		writer.writeheader()
		try:
			query = cls.query_default_order()
		except NotImplementedError:
			query = cls.query
		for obj in query:
			writer.writerow(obj.__dict__)


class IntegerPKey():
	"""Mixin for models with an integer 'id' as the primary key."""

	id = Column(Integer, primary_key=True)

	@classmethod
	def cast(cls, obj, **kwargs):
		if isinstance(obj, int):
			value = cls.query.get(obj)
			if not value:
				raise ValueError(f"No {cls} could be found from int {obj}")
			return value
		else:
			return super().cast(obj, **kwargs)

	@classmethod
	def _get_pkey_col(cls):
		return cls.id

	def __hash__(self):
		return self.id

	def __eq__(self, other):
		return self.__class__ == other.__class__ and self.id == other.id

	@classmethod
	def query_default_order(cls):
		return cls.query.order_by(cls.id)


class AbbrPKey():

	abbr = Column(String, primary_key=True)

	@classmethod
	def cast(cls, obj, **kwargs):
		if isinstance(obj, str):
			value = cls.query.get(obj)
			if not value:
				raise ValueError(f"No {cls} could be found from str {obj}")
			return value
		else:
			return super().cast(obj, **kwargs)

	@classmethod
	def _get_pkey_col(cls):
		return cls.abbr

	def __hash__(self):
		return hash(self.abbr)

	def __eq__(self, other):
		return self.__class__ == other.__class__ and self.abbr == other.abbr

	@classmethod
	def query_default_order(cls):
		return cls.query.order_by(cls.abbr)



import enum


class HTTPStatus(enum.IntEnum):
	CONTINUE = 100
	SWITCHING_PROTOCOLS = 101
	PROCESSING = 102
	EARLY_HINTS = 103

	OK = 200
	CREATED = 201
	ACCEPTED = 202
	NON_AUTHORITATIVE_INFORMATION = 203
	NO_CONTENT = 204
	RESET_CONTENT = 205
	PARTIAL_CONTENT = 206
	MULTI_STATUS = 207
	ALREADY_REPORTED = 208
	IM_USED = 226

	MULTIPLE_CHOICES = 300
	MOVED_PERMANENTLY = 301
	FOUND = 302
	SEE_OTHER = 303
	NOT_MODIFIED = 304
	USE_PROXY = 305
	SWITCH_PROXY = 306
	TEMPORARY_REDIRECT = 307
	PERMANENT_REDIRECT = 308

	BAD_REQUEST = 400
	UNAUTHORIZED = 401
	PAYMENT_REQUIRED = 402
	FORBIDDEN = 403
	NOT_FOUND = 404
	METHOD_NOT_ALLOWED = 405
	NOT_ACCEPTABLE = 406
	PROXY_AUTHENTICATION_REQUIRED = 407
	REQUEST_TIMEOUT = 408
	CONFLICT = 409
	GONE = 410
	LENGTH_REQUIRED = 411
	PRECONDITION_FAILED = 412
	PAYLOAD_TOO_LARGE = 413
	URI_TOO_LONG = 414
	UNSUPPORTED_MEDIA_TYPE = 415
	RANGE_NOT_SATISFIABLE = 416
	EXPECTATION_FAILED = 417
	IM_A_TEAPOT = 418
	MISDIRECTED_REQUEST = 421
	UNPROCESSABLE_ENTITY = 422
	LOCKED = 423
	FAILED_DEPENDENCY = 424
	UPGRADE_REQUIRED = 426
	PRECONDITION_REQUIRED = 428
	TOO_MANY_REQUESTS = 429
	REQUEST_HEADER_FIELDS_TOO_LARGE = 431
	UNAVAILABLE_FOR_LEGAL_REASONS = 451

	INTERNAL_SERVER_ERROR = 500
	NOT_IMPLEMENTED = 501
	BAD_GATEWAY = 502
	SERVICE_UNAVAILABLE = 503
	GATEWAY_TIMEOUT = 504
	HTTP_VERSION_NOT_SUPPORTED = 505
	VARIANT_ALSO_NEGOTIATES = 506
	INSUFFICIENT_STORAGE = 507
	LOOP_DETECTED = 508
	NOT_EXTENDED = 510
	NETWORK_AUTHENTICATION_REQUIRED = 511



"""util - simple functions needed in multiple modules"""
import functools
import urllib.parse


STATE_ABBRS = {
	'Alabama': 'AL',
	'Alaska': 'AK',
	'Arizona': 'AZ',
	'Arkansas': 'AR',
	'California': 'CA',
	'Colorado': 'CO',
	'Connecticut': 'CT',
	'Delaware': 'DE',
	'Florida': 'FL',
	'Georgia': 'GA',
	'Hawaii': 'HI',
	'Idaho': 'ID',
	'Illinois': 'IL',
	'Indiana': 'IN',
	'Iowa': 'IA',
	'Kansas': 'KS',
	'Kentucky': 'KY',
	'Louisiana': 'LA',
	'Maine': 'ME',
	'Maryland': 'MD',
	'Massachusetts': 'MA',
	'Michigan': 'MI',
	'Minnesota': 'MN',
	'Mississippi': 'MS',
	'Missouri': 'MO',
	'Montana': 'MT',
	'Nebraska': 'NE',
	'Nevada': 'NV',
	'New Hampshire': 'NH',
	'New Jersey': 'NJ',
	'New Mexico': 'NM',
	'New York': 'NY',
	'North Carolina': 'NC',
	'North Dakota': 'ND',
	'Ohio': 'OH',
	'Oklahoma': 'OK',
	'Oregon': 'OR',
	'Pennsylvania': 'PA',
	'Rhode Island': 'RI',
	'South Carolina': 'SC',
	'South Dakota': 'SD',
	'Tennessee': 'TN',
	'Texas': 'TX',
	'Utah': 'UT',
	'Vermont': 'VT',
	'Virginia': 'VA',
	'Washington': 'WA',
	'West Virginia': 'WV',
	'Wisconsin': 'WI',
	'Wyoming': 'WY',
	'American Samoa': 'AS',
	'District of Columbia': 'DC',
	'Federated States of Micronesia': 'FM',
	'Guam': 'GU',
	'Marshall Islands': 'MH',
	'Northern Mariana Islands': 'MP',
	'Palau': 'PW',
	'Puerto Rico': 'PR',
	'Virgin Islands': 'VI',
	'Armed Forces Americas': 'AA',
	'Armed Forces Europe': 'AE',
	'Armed Forces Pacific': 'AP',
	}


class no_autoflush(object):
	def __init__(self, session):
		self.session = session
		self.autoflush = session.autoflush

	def __enter__(self):
		self.session.autoflush = False

	def __exit__(self, type, value, traceback):
		self.session.autoflush = self.autoflush


class memoized_method(object):
	"""cache the return value of a method

	This class is meant to be used as a decorator of methods. The return value
	from a given method invocation will be cached on the instance whose method
	was invoked. All arguments passed to a method decorated with memoize must
	be hashable.

	If a memoized method is invoked directly on its class the result will not
	be cached. Instead the method will be invoked like a static method:
	class Obj(object):
		@memoize
		def add_to(self, arg):
			return self + arg
	Obj.add_to(1) # not enough arguments
	Obj.add_to(1, 2) # returns 3, result is not cached

	Based on:
	http://code.activestate.com/recipes/577452-a-memoize-decorator-for-instance-methods/
	"""
	def __init__(self, func):
		self.func = func

	def __get__(self, obj, objtype=None):
		if obj is None:
			return self.func
		return functools.partial(self, obj)

	def __call__(self, *args, **kw):
		obj = args[0]
		try:
			cache = obj.__cache
		except AttributeError:
			cache = obj.__cache = {}
		key = (self.func, args[1:], frozenset(kw.items()))
		try:
			res = cache[key]
		except KeyError:
			res = cache[key] = self.func(*args, **kw)
		except TypeError:
			# This happens if an argument is unhashable
			res = self.func(*args, **kw)
		return res


class classproperty(object):

	def __init__(self, fget):
		self.fget = fget

	def __get__(self, owner_self, owner_cls):
		return self.fget(owner_cls)


def valid_email_address(email):
	if not email or len(email) < 6:
		return False
	try:
		local, domain = email.split('@', 1)
	except ValueError:
		return False
	return local and domain and '.' in domain


def create_url(
	scheme='',
	path='',
	query='',
	fragment='',
	username=None,
	password=None,
	hostname=None,
	port=None,
	):
	"""Like urlib.parse.urlunsplit but build the netloc from parts."""
	if not hostname:
		raise ValueError('Need hostname')
	if port:
		netloc_fmt = '{hostname}:{port}'
	else:
		netloc_fmt = '{hostname}'
	if username:
		if password:
			netloc_fmt = '{username}:{password}@%s' % netloc_fmt
		else:
			netloc_fmt = '{username}@%s' % netloc_fmt
	elif password:
		raise ValueError('Cannot handle password without username')
	netloc = netloc_fmt.format(
		username=username,
		password=password,
		hostname=hostname,
		port=port,
		)
	return urllib.parse.urlunsplit((scheme, netloc, path, query, fragment))



"""Perform database admin tasks including migrations using Alembic.
"""
import logging
import sys

from flask import current_app
from flask.cli import AppGroup
import click

from . import migrate, pg

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

cli = AppGroup(name='db', help='Database Administration')
cli.add_command(migrate.cli)
cli.add_command(pg.cli)


@cli.command()
def create_all():
	"""Create database and objects."""
	logger.info('creating all')
	current_app.extensions['sqlalchemy'].db.create_all()


@cli.command()
@click.confirmation_option(help='Are you sure you want to drop the db?')
def drop_all():
	"""Drop all database objects (drop_all & create_all)."""
	if click.confirm('Are you sure you want to drop all data?'):
		logger.info('dropping all')
		current_app.extensions['sqlalchemy'].db.drop_all()
	else:
		sys.exit(1)


@cli.command()
def reset_all():
	"""Drop and recreate all objects."""
	if click.confirm('Are you sure you want to drop all data?'):
		logger.info('dropping all')
		current_app.extensions['sqlalchemy'].db.drop_all()
		logger.info('creating all')
		current_app.extensions['sqlalchemy'].db.create_all()
	else:
		sys.exit(1)



"""date utils"""
import bisect
import calendar
from contextlib import suppress
from datetime import date, datetime, timedelta
from typing import Tuple

from collections_extended import setlist
from .comparable_mixin import ComparableSameClassMixin


(MON, TUE, WED, THU, FRI, SAT, SUN) = range(7)
weekends = (SAT, SUN)


def week_start(d: datetime, starts_on=SUN) -> date:
	"""Return the first day of the week for the passed date.

	Args:
		d (datetime.datetime or datetime.date): When to get the week start for.
		starts_on: The day of the week that the week starts on.

	Returns:
		A datetime.date object for the first day of the week containing d.
	"""
	if isinstance(d, datetime):
		d = d.date()
	else:
		assert isinstance(d, date)
	days = (d.weekday() + 7 - starts_on) % 7
	d -= timedelta(days=days)
	return d


def workday_diff(  # noqa no way to make this simpler
	start_date: datetime,
	end_date: datetime,
	holidays=None,
	) -> float:
	"""Calculate the number of working days between two dates."""
	assert isinstance(start_date, datetime)
	assert isinstance(end_date, datetime)
	holidays = setlist(sorted(holidays or []))
	if end_date < start_date:
		return -workday_diff(end_date, start_date, holidays=holidays)
	elif end_date == start_date:
		return 0
	# first full day is inclusive
	first_full_day = (start_date + timedelta(days=1)).date()
	# last full day is exclusive
	last_full_day = end_date.date()
	if last_full_day > first_full_day:
		num_full_days = (last_full_day - first_full_day) / timedelta(days=1)
		full_weeks, extra_days = divmod(num_full_days, 7)
		full_weeks = int(full_weeks)
		extra_days = int(extra_days)
		num_full_days -= full_weeks * len(weekends)
		# subtract out any working days that fall in the 'shortened week'
		for d in range(extra_days):
			if (first_full_day + timedelta(d)).weekday() in weekends:
				num_full_days -= 1
		# subtract out any holidays
		start_index = bisect.bisect_left(holidays, start_date.date())
		stop_index = bisect.bisect_right(holidays, end_date.date())
		for d in holidays[start_index:stop_index]:
			if d.weekday() not in weekends and first_full_day <= d < last_full_day:
				num_full_days -= 1
	else:
		num_full_days = 0

	# Calculate partial dats
	def exclude_day(d):
		return d.weekday() in weekends or d in holidays

	partial_days = 0.0
	if end_date.date() == start_date.date():
		if not exclude_day(start_date):
			partial_days = (end_date - start_date) / timedelta(days=1)
	else:
		if not exclude_day(start_date):
			start_date_eod = midnight(first_full_day)
			partial_days += (start_date_eod - start_date) / timedelta(days=1)
		if not exclude_day(end_date):
			end_date_eod = midnight(last_full_day)
			partial_days += (end_date - end_date_eod) / timedelta(days=1)
	return partial_days + num_full_days


def midnight(d: date) -> datetime:
	"""Given a datetime.date, return a datetime.datetime of midnight."""
	if d is None:
		return None
	return datetime(d.year, d.month, d.day)


def past_complete_weeks(
	num_weeks: int,
	today=None,
	week_starts_on=SUN,
	) -> Tuple[datetime, datetime]:
	"""Return a tuple marking the beginning and end of the past number of weeks.

	Args:
		num_weeks (int): The number of weeks
	Returns:
		beg (datetime), end (datetime)
	"""
	if today is None:
		today = date.today()
	end = week_start(today, starts_on=week_starts_on)
	num_days = 7 * num_weeks
	start = end - timedelta(days=num_days)
	return start, end


class Month(ComparableSameClassMixin):

	def __init__(self, year, month):
		if not 0 < month <= 12:
			raise ValueError('Invalid month')
		self.year = int(year)
		self.month = int(month)

	def first(self):
		"""Return a datetime.date object for the first of the month."""
		return self.date(1)

	def first_datetime(self):
		"""Return a datetime.date object for the first of the month."""
		return self.datetime(1)

	def first_string(self):
		return self.first().strftime('%m/%d/%Y')

	def mid(self):
		"""Return a datetime.date object for the 15th of the month."""
		return self.date(15)

	def mid_datetime(self):
		"""Return a datetime.date object for the 15th of the month."""
		return datetime(self.year, self.month, 15)

	def mid_string(self):
		return self.mid().strftime('%m/%d/%Y')

	def last(self):
		return date(self.year, self.month, self.num_days())

	def num_days(self):
		return calendar.monthrange(self.year, self.month)[1]

	def date(self, day):
		"""Return a date for the day of the month.

		num can be a negative number to count back from the end of the month.
		eg. date(-1) is the last day of the month.
		"""
		if day < 0:
			day += self.num_days() + 1
		return date(self.year, self.month, day)

	def datetime(self, day, *args, **kwargs):
		return datetime(self.year, self.month, day, *args, **kwargs)

	def pace(self, d=None):
		"""Return how far through this month d is.

		If d isn't passed use datetime.now()
		"""
		if d is None:
			d = datetime.now()
		if self.month == d.month and self.year == d.year:
			days_into_month = (d - self.first_datetime()) / timedelta(days=1)
			return 1.0 - (days_into_month / self.num_days())
		elif self.year > d.year or (self.year == d.year and self.month > d.month):
			return 1.0
		else:
			return 0.0

	def next(self):
		if self.month == 12:
			return Month(self.year + 1, 1)
		else:
			return Month(self.year, self.month + 1)

	@classmethod
	def from_date(cls, d):
		"""Create a Month from a date(time) object."""
		if not d:
			return None
		return cls(d.year, d.month)

	def _cmp_key(self):
		return self.year, self.month

	def __sub__(self, other):
		"""Return the number of months between two months."""
		return 12 * (self.year - other.year) + self.month - other.month

	@staticmethod
	def iter(start, end=None):
		"""Generate months between passed months.

		Args:
			start: The first month to yield.
			end: The final month to yield. If None, yield infinitely.

		Yields:
			Months between the passed values inclusive.
		"""
		month = start
		while end is None or month <= end:
			yield month
			month = month.next()

	def days(self):
		"""Generate the dates in this month."""
		for i in range(self.num_days()):
			yield self.date(i + 1)


# Parsing

def parse_iso_date_with_colon(date_string):
	"""Parse a date in ISO format with a colon in the timezone.

	For example: 2014-05-01T15:08:00-07:00
	Or if the timezone is missing, parse that:
	For example: 2014-05-01T15:08:00
	"""
	with suppress(ValueError):
		return datetime.strptime(date_string, '%Y-%m-%dT%H:%M:%S')
	with suppress(ValueError):
		return datetime.strptime(date_string, '%Y-%m-%dT%H:%M:%S.%f')
	try:
		last_colon_index = date_string.rindex(':')
	except ValueError:
		raise ValueError(date_string)
	clean = date_string[:last_colon_index] + date_string[last_colon_index + 1:]
	return datetime.strptime(clean, '%Y-%m-%dT%H:%M:%S%z')


def parse_date_missing_zero_padding(
	date_string: str,
	sep='/',
	order='mdy',
	min_year=2000,
	) -> date:
	"""Parse dates without zero padding like 3/1/14.

	If min_year is set, then the year string can be two digits and is parsed as
	the year >= min_year with the last two digits as the passed year.

	Args:
		date_string (str): String to parse
		sep (str): The separator between month, day and year
		order (str): 'm', 'd' and 'y' indicating the order the month, day and year
		min_year (str): For two digit years, use the next year after min_year
	"""
	if len(order) != 3 or set(order) != set('mdy'):
		raise ValueError('Invalid order')
	strings = date_string.split(sep)
	if len(strings) != 3:
		raise ValueError('date_string does not contain 3 values')
	day = int(strings[order.index('d')])
	month = int(strings[order.index('m')])
	year = int(strings[order.index('y')])
	if year < min_year:
		if year < 0 or year >= 100:
			raise ValueError
		year = 100 * (min_year // 100) + year
		if year <= min_year:
			year += 100
	return date(year, month, day)



#!/usr/bin/env python
from __future__ import unicode_literals
from codecs import open
import os
from setuptools import setup, find_packages

# Get the long description from the relevant file
here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
	long_description = f.read()


setup(
	name='lenzm_utils',
	version='0.6.0',
	description='Various utils including Flask projects',
	long_description=long_description,
	author='Michael Lenzen',
	author_email='m.lenzen@gmail.com',
	license='MIT',
	url='https://github.com/mlenzen/lenzm_utils',
	packages=find_packages(exclude=('tests*', 'docs', 'examples')),
	include_package_data=True,
	zip_safe=False,
	package_data={
		'': ['README.rst', 'LICENSE'],
		},
	install_requires=[
		'setuptools',
		'Flask>=0.11',
		'SQLAlchemy',
		'Flask-SQLAlchemy',
		'alembic',
		'click',
		'collections_extended',
		'openpyxl',
		'pytz',
		],
	tests_require=[
		'pytest',
		],
	# See: http://pypi.python.org/pypi?%3Aaction=list_classifiers
	classifiers=[
		'Development Status :: 3 - Alpha',
		'Intended Audience :: Developers',
		'Topic :: Software Development',
		'License :: OSI Approved :: BSD License',
		'Programming Language :: Python :: 3',
		'Programming Language :: Python :: 3.6',
		],
	)
